//	Maze4DContentView.swift
//
//	© 2023 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


enum PanelType {
	case noPanel
	case newMazePanel
	case shearPanel
	case exportPanel
	case helpPanel
	case contactPanel
	case translatorsPanel
}


struct Maze4DContentView: View {

	//	"@State" should be used with structs.
	//		Each @State instance is initialized once.
	//		The runtime then makes a copy of
	//		that pre-initialized instance
	//		for each View that gets created.
	//
	//	"@StateObject" should be used with classes.
	//		Each View get its own instance of the class.
	//
	//	Using "@State" with a class would be a bad idea,
	//	because a single instance of it would get initialized,
	//	and then each View would merely get a copy
	//	of the reference to that one instance.
	//	In effect, the different Views (even Views
	//	in different Scenes) would all share the same instance.

	@StateObject var itsModelData = Maze4DModel()
	@State var itsActivePanel = PanelType.newMazePanel

	//	To make App Store screenshots, set gMakeAppStoreScreenshots = true.
	@State var itsScreenshotIndex: Int	//	used only when making screenshots
		= (gLanguageIndependentScreenshots ? 0 : 2)

	//	A renderer is a potentially expensive resource
	//	(it may have to load textures, etc.), so let's let
	//	the GeometryGamesView and the GeometryGamesExportView
	//	share a single renderer, instead of forcing each
	//	to create its own.
	//
	//		Note #1  The renderer's init() may fail,
	//		and @StateObject doesn't seem willing to hold an optional
	//		(because an optional doesn't conform to ObservableObject),
	//		so I've wrapped the Maze4DRenderer
	//		in an OptionalMaze4DRenderer as a workaround.
	//
	//		Note #2  itsOptionalRenderer isn't really an observed state,
	//		it's just a resource that's shared among our subviews.
	//		(But it's not shared between Scenes -- each Scene
	//		gets its own OptionalMaze4DRenderer instance.)
	//
	class OptionalMaze4DRenderer: ObservableObject {
		var itsRenderer: Maze4DRenderer?
		init(_ renderer: Maze4DRenderer?) {
			itsRenderer = renderer
		}
	}
	@StateObject var itsOptionalRenderer
		= OptionalMaze4DRenderer(Maze4DRenderer())

	//	State variables for use in the gesture handlers.
	@State var dragPurpose: DragPurpose = .unknown
	@State var previousSliderTouchIntrinsic: SIMD2<Double> = .zero
	@State var itsRecentPoints: [RecentPoint] = []
	@State var itsPreviousAngle: RecentAngle? = nil


	var body: some View {
	
		let theTranslatorThanks = [
			GeometryGamesTranslatorThanks(pre: "mit Dank an", name: "Bernd Macheleidt und\nSascha Gritzbach", post: nil, lang: "de"),
			GeometryGamesTranslatorThanks(pre: "grazie a", name: "Daniela Bigatti", post: nil, lang: "it"),
			GeometryGamesTranslatorThanks(pre: nil, name: "竹内建氏に", post: "感謝します", lang: "ja"),
			GeometryGamesTranslatorThanks(pre: "z podziękowaniami dla", name: "Marcina Pilarskiego", post: nil, lang: "pl"),
			GeometryGamesTranslatorThanks(pre: "obrigado a", name: "Nuno Fernandes", post: nil, lang: "pt"),
			GeometryGamesTranslatorThanks(pre: "спасибо", name: "Маше Трнковой", post: nil, lang: "ru"),
			GeometryGamesTranslatorThanks(pre: nil, name: "Süleyman Saltuk Güngör'e", post: "teşekkürler", lang: "tr")
		]

		ZStack(alignment: .bottom) {
		
			//	Main animation view
			if let theRenderer = itsOptionalRenderer.itsRenderer {

				GeometryReader() { geometryReaderProxy in
				
					GeometryGamesViewRep(
						modelData: itsModelData,
						renderer: theRenderer
					)
					//
					//	The order in which the gestures get attached is important:
					//
					//		.gesture(DragGesture()…)
					//		.gesture(RotationGesture()…)
					//
					//	works fine, but
					//
					//		.gesture(RotationGesture()…)
					//		.gesture(DragGesture()…)
					//
					//	fails to properly handle the DragGesture.
					//	I can't find any documentation about this
					//	(nor even any blog posts or other discussion online),
					//	but I'm _guessing_ that the explanation for this behavior
					//	is the following:  In the first case shown above,
					//	as soon as the user places two fingers on the display,
					//	SwiftUI knows the DragGesture has failed and moves on
					//	to consider the RotationGesture.  But in the second case,
					//	when the user drags a finger on the display, SwiftUI
					//	doesn't give up hope for the RotationGesture,
					//	so the DragGesture receives no onChanged messages
					//	(even though it does eventually receive a final onEnded message).
					//
					.gesture(maze4DDragGesture(
								modelData: itsModelData,
								viewSize: geometryReaderProxy.size,
								dragPurpose: $dragPurpose,
								previousSliderTouchIntrinsic: $previousSliderTouchIntrinsic,
								recentPoints: $itsRecentPoints))
					.gesture(maze4DRotationGesture(
								modelData: itsModelData,
								previousAngle: $itsPreviousAngle,
								recentPoints: $itsRecentPoints))
					.simultaneousGesture(
						//	If we made a TapGesture the top-priority gesture,
						//	the user would have to wait for it to fail before
						//	getting a response to a DragGesture.  So let's test
						//	for a TapGesture in parallel with the Drag/Rotate
						//	gestures, and use it only to dismiss an open panel.
						TapGesture()
						.onEnded() { _ in itsActivePanel = .noPanel })
				}
				.edgesIgnoringSafeArea(.all)
			}
			
			VStack() {
			
				//	The active panel, if any
				Group() {
				
					//	Note:  Some of the panels get leading-aligned, others trailing-aligned.
					switch itsActivePanel {
					
					case .newMazePanel:
						if let theRenderer = itsOptionalRenderer.itsRenderer {
							NewMazeView(
								modelData: itsModelData,
								activePanel: $itsActivePanel,
								renderer: theRenderer)
							.frame(maxWidth: .infinity, alignment: .leading)
						}

					case .shearPanel:
						if let theRenderer = itsOptionalRenderer.itsRenderer {
							ShearView(
								modelData: itsModelData,
								renderer: theRenderer)
							.frame(maxWidth: .infinity, alignment: .trailing)
						}

					case .exportPanel:
						if let theRenderer = itsOptionalRenderer.itsRenderer {
							GeometryGamesExportView(
								modelData: itsModelData,
								exportRenderer: theRenderer)
							.preferredColorScheme(.light)
							.frame(maxWidth: .infinity, alignment: .trailing)
						}
					
					case .helpPanel:
						Maze4DHelpView()
						.frame(maxWidth: .infinity, alignment: .trailing)

					case .contactPanel:
						GeometryGamesContactView()
						.preferredColorScheme(.light)
						.frame(maxWidth: .infinity, alignment: .trailing)

					case .translatorsPanel:
						GeometryGamesTranslatorsView(theTranslatorThanks)
						.frame(maxWidth: .infinity, alignment: .trailing)

					default:
						Spacer()
					}
				}
			
				//	The buttons
				HStack(alignment: .bottom) {
					NewMazeButton(activePanel: $itsActivePanel)
					Spacer()
					ShearButton(activePanel: $itsActivePanel)
					ExportButton(activePanel: $itsActivePanel)
					HelpButton(activePanel: $itsActivePanel)
				}
			}
			.padding(geometryGamesControlInset)
				
			if gMakeAppStoreScreenshots && !gGetScreenshotOrientations {
				//	The Rectangle will respond to taps
				//	only if it's at least a tiny bit opaque.
				Rectangle()
				.foregroundColor(Color(.displayP3,
									red: 0.0,
									green: 0.0,
									blue: 0.0,
									opacity: 0.000000000001))
				.edgesIgnoringSafeArea(.all)
				.onTapGesture {
					switch itsScreenshotIndex {
					case 0: itsScreenshotIndex = 1	//	toggle between language-independent screenshots
					case 1: itsScreenshotIndex = 0
					case 2: itsScreenshotIndex = 3	//	toggle between language-dependent screenshots
					case 3: itsScreenshotIndex = 2
					default: assertionFailure("Unexpected value of itsScreenshotIndex")
					}
				}
			}
		}
		.onAppear() {
			if gMakeAppStoreScreenshots {
				prepareForAppStoreScreenshot(
					itsScreenshotIndex,
					modelData: itsModelData,
					activePanel: $itsActivePanel,
					optionalRenderer: itsOptionalRenderer.itsRenderer)
			}
		}
		.onChange(of: itsScreenshotIndex) { aNewScreenshotIndex in
			prepareForAppStoreScreenshot(
				aNewScreenshotIndex,
				modelData: itsModelData,
				activePanel: $itsActivePanel,
				optionalRenderer: itsOptionalRenderer.itsRenderer)
		}
	}
}


struct NewMazeButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .newMazePanel
		} label: {
			Image(systemName: "arrow.clockwise")
				.font(.title)
				.padding(geometryGamesControlPadding)
		}
	}
}

struct NewMazeView: View {

	@ObservedObject var modelData: Maze4DModel
	@Binding var activePanel: PanelType
	let renderer: Maze4DRenderer

	var body: some View {

		VStack(alignment: .leading) {

			ForEach(DifficultyLevel.allCases, id: \.id) { difficultyLevel in

				Button(
					action: {
						modelData.itsMaze = makeNewMaze(difficultyLevel)
						renderer.refreshForNewMaze(modelData: modelData)
						activePanel = .noPanel
					},
					label: {

						let n = difficultyLevel.n
								
						HStack(spacing: 16.0) {

							Circle()
								.fill(difficultyLevel.color)
								.frame(width: 32.0, height: 32.0)

							VStack(alignment: .leading) {
								Text(difficultyLevel.localizedNameKey)
									.lineLimit(1)

								Text(verbatim: "\(n) × \(n) × \(n) × \(n)")
									.foregroundColor(.gray)
									.font(.footnote)
							}
						}
					}
				)
			}
		}
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.preferredColorScheme(.light)
		.cornerRadius(geometryGamesCornerRadius)
	}
}


struct ShearButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .shearPanel
		} label: {
			Image(systemName: "arrow.left.arrow.right.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
		}
	}
}

struct ShearView: View {

	@ObservedObject var modelData: Maze4DModel
	let renderer: Maze4DRenderer

	var body: some View {

		VStack() {
			Text("4D Shear")
			Slider(
				value: $modelData.its4DShearFactor,
				in: min4DShearFactor ... max4DShearFactor
			)
			.onChange(of: modelData.its4DShearFactor) { aNewShearFactor in
				renderer.refreshForNewShear(modelData: modelData)
			}
		}
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.preferredColorScheme(.light)
		.cornerRadius(geometryGamesCornerRadius)
		.frame(width: 256.0)
	}
}


struct ExportButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .exportPanel
		} label: {
			Image(systemName: "square.and.arrow.up")
				.font(.title)
				.padding(geometryGamesControlPadding)
		}
	}
}


struct HelpButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Menu() {

			Button() {
				activePanel = .helpPanel
			} label: {
				Label("4D Maze Help", systemImage: "questionmark.circle")
			}

			Button() {
				activePanel = .contactPanel
			} label: {
				GeometryGamesHelpMenuContactItemLabel()
			}

			Button() {
				activePanel = .translatorsPanel
			} label: {
				GeometryGamesHelpMenuTranslatorsItemLabel()
			}

		} label: {

			Image(systemName: "questionmark.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
		}
	}
}

let headerColor = Color(.displayP3, red: 0.0, green: 0.25, blue: 0.5)

//	Avoid italics in non-European languages.
let gFontDesign = {
	let theLanguage = Locale.preferredLanguages.first ?? "en"
	switch theLanguage {
		case "ja": return Font.Design.default
		default:   return Font.Design.serif
	}
}()

struct Maze4DHelpView: View {

	var body: some View {

		return ScrollView {
			VStack(alignment: .leading, spacing: 16) {

				Text("4D Maze")
					.font(.title2)
					.foregroundColor(headerColor)
					.padding(EdgeInsets(top: 16.0, leading: 0.0, bottom: 0.0, trailing: 0.0))
					.frame(maxWidth: .infinity, alignment: .center)

				Maze4DHelpSection(title: "HowToPlay-title") {
				
					Text("HowToPlay-text")
						.modifier(Maze4DBodyText())
					Image("Help/SliderToBox")
						.resizable()
						.aspectRatio(contentMode: .fit)
						.frame(width: 192.0)
				}

				Maze4DHelpSection(title: "AnaKata-title") {
				
					Text("AnaKata-text")
						.modifier(Maze4DBodyText())
					Spacer()
						.frame(height: 4.0)
					HStack(spacing: 0.0) {
						VStack(alignment: .trailing) {
							Text("AnaKata-label1")
							Text("AnaKata-label2")
							Text("AnaKata-label3")
							Text("AnaKata-label4")
						}
						Spacer()
							.frame(width: 8.0)
						VStack(alignment: .trailing) {
							Text("AnaKata-pos1")
							Text("AnaKata-pos2")
							Text("AnaKata-pos3")
							Text("AnaKata-pos4")
						}
						Spacer()
							.frame(width: 2.0)
						VStack() {
							Text(verbatim: "/")
							Text(verbatim: "/")
							Text(verbatim: "/")
							Text(verbatim: "/")
						}
						Spacer()
							.frame(width: 2.0)
						VStack(alignment: .leading) {
							Text("AnaKata-neg1")
							Text("AnaKata-neg2")
							Text("AnaKata-neg3")
							Text("AnaKata-neg4")
						}
					}
					.font(.system(size: 15.0, design: gFontDesign))
				}

				Maze4DHelpSection(title: "Colors-title") {
				
					Text("Colors-text1")
						.modifier(Maze4DBodyText())
					Text("Colors-caption1")
						.modifier(Maze4DCaptionText())
					Image("Help/ColoredTubes")
						.resizable()
						.aspectRatio(contentMode: .fit)
						.frame(width: 256.0)
						
					Spacer()
						.frame(height: 16.0)
						
					Text("Colors-text2")
						.modifier(Maze4DBodyText())
					Text("Colors-caption2")
						.modifier(Maze4DCaptionText())
					Image("Help/RainbowTube")
						.resizable()
						.aspectRatio(contentMode: .fit)
						.frame(width: 56.0)
				}

				Maze4DHelpSection(title: "SuperimposedPoints-title") {
				
					Text("SuperimposedPoints-text")
						.modifier(Maze4DBodyText())
				}

				Maze4DHelpSection(title: "4DConnections-title") {
				
					Text("4DConnections-text")
						.modifier(Maze4DBodyText())

					//	Even a tiny bit of top padding on the HStack
					//	seems to add a whole line of empty space.
					//	So let's use a Spacer instead.
					Spacer()
						.frame(height: 8.0)

					//	Side-by-side captions
					HStack(alignment: .center, spacing: 8.0) {
						Text("4DConnections-caption1")
							.multilineTextAlignment(.center)
							.font(.caption)
							.frame(width: 128.0)
						Text("4DConnections-caption2")
							.multilineTextAlignment(.center)
							.font(.caption)
							.frame(width: 128.0)
					}

					//	Side-by-side figures
					HStack(alignment: .center, spacing: 8.0) {
						Image("Help/ConnectedTubes")
							.resizable()
							.aspectRatio(contentMode: .fit)
							.frame(width: 128.0)
						Image("Help/UnconnectedTubes")
							.resizable()
							.aspectRatio(contentMode: .fit)
							.frame(width: 128.0)
					}
				}
				
				Text("Questions-footnote")
					.font(.system(.footnote, design: gFontDesign))
				
				Text(verbatim: "© 2023 by Jeff Weeks")
					.font(.system(.footnote, design: gFontDesign))
			}
		}
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.preferredColorScheme(.light)
		.cornerRadius(geometryGamesCornerRadius)
		.frame(width: 352.0)
		.frame(maxHeight: .infinity)
	}
}

struct Maze4DHelpSection<Content: View>: View {
	
	let title: LocalizedStringKey
	let content: Content
	
	init(
		title: LocalizedStringKey,
		@ViewBuilder content: () -> Content
	) {
	
		self.title = title
		self.content = content()
	}

	var body: some View {

		VStack(alignment: .center) {

			Text(title)
				.font(.title3)
				.foregroundColor(headerColor)
				.frame(maxWidth: .infinity, alignment: .leading)

			Spacer()
				.frame(height: 6.0)
			
			content
		}
	}
}

struct Maze4DBodyText: ViewModifier {

	func body(content: Content) -> some View {

		content
			.font(.system(.body, design: .serif))
			.padding(EdgeInsets(top: 0.0, leading: 16.0, bottom: 0.0, trailing: 0.0))
			.frame(maxWidth: .infinity, alignment: .leading)
	}
}

struct Maze4DCaptionText: ViewModifier {

	func body(content: Content) -> some View {

		//	Even a tiny bit of top padding on the content
		//	seems to add a whole line of empty space.
		//	So let's set the top padding to zero
		//	and use a Spacer instead.
		Spacer()
			.frame(height: 8.0)
			
		content
			.multilineTextAlignment(.center)
			.font(.caption)
			.padding(EdgeInsets(top: 0.0, leading: 20.0, bottom: 0.0, trailing: 20.0))
	}
}
