//	Maze4DModel.swift
//
//	© 2023 by Jeff Weeks
//	See TermsOfUse.txt

import Foundation


let min4DShearFactor = 0.125	//	must be greater than zero
let max4DShearFactor = 2.0 * sqrt(2.0)
let default4DShearFactor = 1.375 * sqrt(2.0)	//	√2 => tubes are tangent

//	Enable gMaze4DForTalks only briefly,
//	when compiling 4D Maze for use on macOS during talks,
//	and then disable it again.
//#warning("disable gMaze4DForTalks")
let gMaze4DForTalks = false


class Maze4DModel: ObservableObject, ModelIsAnimatable {
				//	Declaring conformance to ObservableObject
				//	lets Maze4DContentView use a Maze4DModel as a @StateObject

	var itsMaze: MazeData? = nil {
		didSet {
			itsGameIsOver = false
			itsDisplayChangeCount += 1
		}
	}

	var itsOrientation: simd_quatd = gQuaternionIdentity {
		didSet {
			itsDisplayChangeCount += 1			
		}
	}
	
	//	Assuming a constant frame rate, with gFramePeriod seconds
	//	from the start of one frame to the start of the next,
	//	let itsIncrement be the small additional rotation
	//	that we want post-multiply for each successive frame.
	//
	var itsIncrement: simd_quatd?	//	= nil when maze isn't rotating

	//	To distinguish nodes whose x-, y- and z-coordinates all agree,
	//	we slightly shear the 4D space before projecting down to 3D.
	//
	//	its4DShearFactor determines the relative spacing between tubes whose
	//	positions differ only in the ana-kata dimension.  The UI lets the player
	//	vary its4DShearFactor, so s/he can make it larger for good visibility,
	//	or make it smaller to better understand that parallel edges differing
	//	only in the ana-kata dimension really should appear superimposed in 3D.
	//	A reasonable default value is √2.  Of course its4DShearFactor should
	//	never be exactly 0, partly to avoid possible division by 0, but mainly
	//	to avoid the z-fighting that would occur if we tried to superimpose
	//	two tubes exactly.
	//
	//		Note:  its4DShearFactor is @Published because
	//		whenever it changes we'll need to replace
	//		the gestures that depend on it.
	//
	@Published var its4DShearFactor: Double =
		(gMaze4DForTalks ?
			min4DShearFactor :
			gMakeAppStoreScreenshots ?
				1.0 :
				default4DShearFactor)
	
	var itsGameIsOver: Bool = false {
		didSet {
			if itsGameIsOver {
				enqueueSoundRequest("MazeSolved.m4a")
			}
			itsDisplayChangeCount += 1
		}
	}
	
	//	To help new users understand that they should drag the slider,
	//	we'll let the slider blink when users launch the app.
	//	But as soon as the user touches the slider it will stop
	//	blinking and won't blink again until the user quits the app
	//	and restarts it.  At the moment it stops blinking, the goal
	//	will flash briefly to let the user know where to drag
	//	the slider too.
	//
	//	As a practical matter, the following two variables will
	//	record when the slider started blinking and when the goal
	//	began its flash.  When they aren't blinking or flashing,
	//	these variables will be nil.
	var itsSliderBlinkStartTime: CFAbsoluteTime?
		= (gMakeAppStoreScreenshots ?
			nil :						//	Doesn't blink for screenshots.
			CFAbsoluteTimeGetCurrent()	//	Blinks at launch.
		  )
	var itsGoalFlashStartTime: CFAbsoluteTime? = nil {	//	doesn't flash until later
		didSet {
			itsDisplayChangeCount += 1
		}
	}

	var itsDisplayChangeCount: UInt64 = 0

	func updateModel() -> UInt64 {	//	returns the model's display change count

		if let theIncrement = itsIncrement {

			//	Post-multiply ( => left-multiply) itsOrientation by theIncrement.
			itsOrientation = theIncrement * itsOrientation

			//	Normalize itsOrientation to keep its length
			//	from gradually drifting away from 1.
			//
			//		Note:  The length stays very close to 1
			//		at all times, so we don't have to worry about
			//		passing in the zero quaternion here.
			//
			itsOrientation = simd_normalize(itsOrientation)

			itsDisplayChangeCount += 1
		}
		
		//	Redraw the scene continuously while the slider is blinking.
		if itsSliderBlinkStartTime != nil {
			itsDisplayChangeCount += 1
		}
		
		//	We don't need to continuously redraw the highlighted goal,
		//	but we do need to un-highlight it after a fraction of a second.
		if let theGoalFlashStartTime = itsGoalFlashStartTime {
			let theElapsedTime = CFAbsoluteTimeGetCurrent() - theGoalFlashStartTime
			let theFlashDuration = 0.25	//	seconds
			if theElapsedTime > theFlashDuration {
				itsGoalFlashStartTime = nil
			}
		}

		return itsDisplayChangeCount
	}
}
